@@ -95,10 +95,44 @@ ILCallFrame *_ILAllocCallFrame(ILExecThread *thread)
 #endif
 }
 
+
+#define CHECK_MANAGED_BARRIER()	\
+	if (thread->managedSafePointFlags)	\
+	{	\
+		if  ((thread->managedSafePointFlags & _IL_MANAGED_SAFEPOINT_THREAD_ABORT) && ILThreadIsAbortRequested()) \
+		{ \
+			if (_ILExecThreadSelfAborting(thread) == 0) \
+			{ \
+				ILThreadAtomicStart(); \
+				thread->managedSafePointFlags &= ~_IL_MANAGED_SAFEPOINT_THREAD_ABORT; \
+				ILThreadAtomicEnd(); \
+				stacktop[0].ptrValue = thread->thrownException; \
+				thread->thrownException = 0; \
+				stacktop += 1; \
+				goto throwException; \
+			} \
+		} \
+		else if (thread->managedSafePointFlags & _IL_MANAGED_SAFEPOINT_THREAD_SUSPEND) \
+		{ \
+			ILThreadAtomicStart(); \
+			thread->managedSafePointFlags &= ~_IL_MANAGED_SAFEPOINT_THREAD_SUSPEND; \
+			ILThreadAtomicEnd(); \
+			_ILExecThreadSuspendThread(thread, thread->supportThread); \
+		} \
+	}
+
+#define BEGIN_NATIVE_CALL()	\
+	thread->runningManagedCode = 0;
+
+#define END_NATIVE_CALL() \
+	CHECK_MANAGED_BARRIER(); \
+	thread->runningManagedCode = 1;
+
 #ifdef IL_DUMP_CVM
 #define	DUMP_STACK()	\
 			do { \
 				int posn; \
+				BEGIN_NATIVE_CALL(); \
 				fprintf(IL_DUMP_CVM_STREAM, "Stack:"); \
 				for(posn = 1; posn <= 16; ++posn) \
 				{ \
@@ -110,14 +144,17 @@ ILCallFrame *_ILAllocCallFrame(ILExecThread *thread)
 				} \
 				putc('\n', IL_DUMP_CVM_STREAM); \
 				fflush(IL_DUMP_CVM_STREAM); \
+				END_NATIVE_CALL(); \
 			} while (0)
 #define	REPORT_METHOD_CALL()	\
 			do { \
+				BEGIN_NATIVE_CALL(); \
 				fprintf(IL_DUMP_CVM_STREAM, "Entering %s::%s (%ld)\n", \
 					    methodToCall->member.owner->className->name, \
 					    methodToCall->member.name, \
 					    (long)(stacktop - thread->stackBase)); \
 				DUMP_STACK(); \
+				END_NATIVE_CALL(); \
 			} while (0)
 #else
 #define	REPORT_METHOD_CALL()
@@ -131,7 +168,9 @@ ILCallFrame *_ILAllocCallFrame(ILExecThread *thread)
 				} \
 				else \
 				{ \
+					BEGIN_NATIVE_CALL(); \
 					callFrame = _ILAllocCallFrame(thread); \
+					END_NATIVE_CALL(); \
 					if(!callFrame) \
 					{ \
 						STACK_OVERFLOW_EXCEPTION(); \
@@ -158,39 +197,6 @@ ILCallFrame *_ILAllocCallFrame(ILExecThread *thread)
 				} \
 			} while (0)
 
-#define CHECK_ABORT()	\
-	if (thread->abortRequested && ILThreadIsAbortRequested())	\
-	{	\
-		thread->abortRequested = 0;	\
-		/* Mark the thread as aborted if it hasn't already been aborted	*/ \
-		if (ILThreadSelfAborting(thread->supportThread))	\
-		{	\
-			/* Great, thread hasn't aborted itself yet so throw a ThreadAbortException. */ \
-			stacktop[0].ptrValue = _ILExecThreadNewThreadAbortException(thread, 0); \
-			thread->aborting = 1; \
-			thread->abortHandlerEndPC = 0; \
-			thread->abortHandlerFrame = 0; \
-			thread->threadAbortException = 0; \
-			stacktop += 1; \
-			thread->runningManagedCode = 1;	\
-			if (!ILCoderPCToHandler(thread->process->coder, pc, 0)) \
-			{ \
-				goto throwException; \
-			} \
-			else \
-			{ \
-				goto prefixThrowException; \
-			} \
-		}	\
-	}
-
-#define BEGIN_NATIVE_CALL()	\
-	thread->runningManagedCode = 0;
-
-#define END_NATIVE_CALL() \
-	CHECK_ABORT(); \
-	thread->runningManagedCode = 1;
-
 /*
  * Determine the number of stack words that are occupied
  * by a specific type.
@@ -617,13 +623,19 @@ VMCASE(COP_CALL):
 		/* Copy the state back into the thread object */
 		COPY_STATE_TO_THREAD();
 
+		BEGIN_NATIVE_CALL();
+
 		/* Convert the method */
 		tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 		if(!tempptr)
 		{
+			END_NATIVE_CALL();
+
 			CONVERT_FAILED_EXCEPTION();
 		}
 
+		END_NATIVE_CALL();
+
 		/* Allocate a new call frame */
 		ALLOC_CALL_FRAME();
 
@@ -702,12 +714,17 @@ VMCASE(COP_CALL_CTOR):
 		COPY_STATE_TO_THREAD();
 
 		/* Convert the method */
+		BEGIN_NATIVE_CALL();
+
 		tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 		if(!tempptr)
 		{
+			END_NATIVE_CALL();
 			CONVERT_FAILED_EXCEPTION();
 		}
 
+		END_NATIVE_CALL();
+
 		/* Allocate a new call frame */
 		ALLOC_CALL_FRAME();
 
@@ -911,7 +928,7 @@ VMCASE(COP_CALL_VIRTUAL):
 {
 	/* Call a virtual method */
 	tempptr = stacktop[-((ILInt32)CVM_ARG_DWIDE1_SMALL)].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		/* Locate the method to be called */
 		methodToCall = (GetObjectClassPrivate(tempptr))
@@ -941,12 +958,18 @@ VMCASE(COP_CALL_VIRTUAL):
 			COPY_STATE_TO_THREAD();
 
 			/* Convert the method */
+			BEGIN_NATIVE_CALL();
+
 			tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 			if(!tempptr)
 			{
+				END_NATIVE_CALL();
+
 				CONVERT_FAILED_EXCEPTION();
 			}
 
+			END_NATIVE_CALL();
+
 			/* Allocate a new call frame */
 			ALLOC_CALL_FRAME();
 
@@ -964,10 +987,7 @@ VMCASE(COP_CALL_VIRTUAL):
 			CVM_OPTIMIZE_BLOCK();
 		}
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAK(COP_CALL_VIRTUAL);
 
@@ -1007,7 +1027,7 @@ VMCASE(COP_CALL_INTERFACE):
 {
 	/* Call an interface method */
 	tempptr = stacktop[-((ILInt32)CVM_ARG_DWIDE1_SMALL)].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		/* Locate the method to be called */
 	#ifdef IL_USE_IMTS
@@ -1058,11 +1078,17 @@ VMCASE(COP_CALL_INTERFACE):
 			COPY_STATE_TO_THREAD();
 	
 			/* Convert the method */
+			BEGIN_NATIVE_CALL();
+
 			tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 			if(!tempptr)
 			{
+				END_NATIVE_CALL();
+
 				CONVERT_FAILED_EXCEPTION();
 			}
+			
+			END_NATIVE_CALL();
 
 			/* Allocate a new call frame */
 			ALLOC_CALL_FRAME();
@@ -1081,10 +1107,7 @@ VMCASE(COP_CALL_INTERFACE):
 			CVM_OPTIMIZE_BLOCK();
 		}
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAK(COP_CALL_INTERFACE);
 
@@ -1155,7 +1178,7 @@ VMCASE(COP_RETURN):
 	stacktop = frame;
 popFrame:
 
-	CHECK_ABORT();
+	CHECK_MANAGED_BARRIER();
 
 	callFrame = &(thread->frameStack[--(thread->numFrames)]);
 	methodToCall = callFrame->method;
@@ -1167,6 +1190,9 @@ popFrame:
 	/* Should we return to an external method? */
 	if(pc == IL_INVALID_PC)
 	{
+#if defined(IL_USE_INTERRUPT_BASED_NULL_POINTER_CHECKS)
+		IL_MEMCPY(&thread->exceptionJumpBuffer, &backupJumpBuffer, sizeof(IL_JMP_BUFFER));
+#endif
 		COPY_STATE_TO_THREAD();
 		return 0;
 	}
@@ -1465,11 +1491,17 @@ VMCASE(COP_CALLI):
 		COPY_STATE_TO_THREAD();
 
 		/* Convert the method */
+		BEGIN_NATIVE_CALL();
+
 		tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 		if(!tempptr)
 		{
+			END_NATIVE_CALL();
+
 			CONVERT_FAILED_EXCEPTION();
 		}
+		
+		END_NATIVE_CALL();
 
 		/* Allocate a new call frame */
 		ALLOC_CALL_FRAME();
@@ -1508,7 +1540,7 @@ case COP_CALL_VIRTUAL:
 {
 	/* Wide version of "call_virtual" */
 	tempptr = stacktop[-((ILInt32)CVM_ARG_DWIDE1_LARGE)].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		/* Locate the method to be called */
 		methodToCall = (GetObjectClassPrivate(tempptr))
@@ -1518,11 +1550,17 @@ case COP_CALL_VIRTUAL:
 		COPY_STATE_TO_THREAD();
 
 		/* Convert the method */
+		BEGIN_NATIVE_CALL();
+
 		tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 		if(!tempptr)
 		{
+			END_NATIVE_CALL();
+
 			CONVERT_FAILED_EXCEPTION();
 		}
+		
+		END_NATIVE_CALL();
 
 		/* Allocate a new call frame */
 		ALLOC_CALL_FRAME();
@@ -1540,10 +1578,7 @@ case COP_CALL_VIRTUAL:
 		method = methodToCall;
 		CVM_OPTIMIZE_BLOCK();
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAKNOEND;
 
@@ -1551,7 +1586,7 @@ case COP_CALL_INTERFACE:
 {
 	/* Wide version of "call_interface" */
 	tempptr = stacktop[-((ILInt32)CVM_ARG_DWIDE1_LARGE)].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		/* Locate the method to be called */
 	#ifdef IL_USE_IMTS
@@ -1582,11 +1617,17 @@ case COP_CALL_INTERFACE:
 		COPY_STATE_TO_THREAD();
 
 		/* Convert the method */
+		BEGIN_NATIVE_CALL();
+
 		tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 		if(!tempptr)
 		{
+			END_NATIVE_CALL();
+
 			CONVERT_FAILED_EXCEPTION();
 		}
+		
+		END_NATIVE_CALL();
 
 		/* Allocate a new call frame */
 		ALLOC_CALL_FRAME();
@@ -1604,10 +1645,7 @@ case COP_CALL_INTERFACE:
 		method = methodToCall;
 		CVM_OPTIMIZE_BLOCK();
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAKNOEND;
 
@@ -1660,11 +1698,18 @@ performTailCall:
 	else
 	{
 		COPY_STATE_TO_THREAD();
+		BEGIN_NATIVE_CALL();
+
 		tempptr = (void *)(_ILConvertMethod(thread, methodToCall));
 		if (!tempptr)
 		{
+			END_NATIVE_CALL();
+
 			CONVERT_FAILED_EXCEPTION();
 		}
+		
+		END_NATIVE_CALL();
+
 		RESTORE_STATE_FROM_THREAD();
 	}
 
@@ -1730,16 +1775,13 @@ VMCASE(COP_PREFIX_TAIL_CALLVIRT):
 {
 	/* Call a virtual method */
 	tempptr = stacktop[-((ILInt32)CVMP_ARG_WORD)].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		/* Locate the method to be called */
 		methodToCall = (GetObjectClassPrivate(tempptr))->vtable[CVMP_ARG_WORD2];
 		goto performTailCall;
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAK(COP_PREFIX_TAIL_CALLVIRT);
 
@@ -1761,7 +1803,7 @@ VMCASE(COP_PREFIX_TAIL_CALLINTF):
 {
 	/* Call an interface method */
 	tempptr = stacktop[-((ILInt32)CVMP_ARG_WORD)].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		/* Locate the method to be called */
 	#ifdef IL_USE_IMTS
@@ -1788,10 +1830,7 @@ VMCASE(COP_PREFIX_TAIL_CALLINTF):
 	#endif
 		goto performTailCall;
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAK(COP_PREFIX_TAIL_CALLINTF);
 
@@ -1845,16 +1884,13 @@ VMCASE(COP_PREFIX_LDVIRTFTN):
 {
 	/* Load the address of a virtual function onto the stack */
 	tempptr = stacktop[-1].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		stacktop[-1].ptrValue =
 			(GetObjectClassPrivate(tempptr))->vtable[CVMP_ARG_WORD];
 		MODIFY_PC_AND_STACK(CVMP_LEN_WORD, 0);
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAK(COP_PREFIX_LDVIRTFTN);
 
@@ -1885,7 +1921,7 @@ VMCASE(COP_PREFIX_LDINTERFFTN):
 {
 	/* Load the address of an interface function onto the stack */
 	tempptr = stacktop[-1].ptrValue;
-	if(tempptr)
+	BEGIN_NULL_CHECK(tempptr)
 	{
 		stacktop[-1].ptrValue =
 			_ILLookupInterfaceMethod(GetObjectClassPrivate(tempptr),
@@ -1893,10 +1929,7 @@ VMCASE(COP_PREFIX_LDINTERFFTN):
 									 CVMP_ARG_WORD);
 		MODIFY_PC_AND_STACK(CVMP_LEN_WORD_PTR, 0);
 	}
-	else
-	{
-		NULL_POINTER_EXCEPTION();
-	}
+	END_NULL_CHECK();
 }
 VMBREAK(COP_PREFIX_LDINTERFFTN);
 
@@ -1934,11 +1967,13 @@ VMBREAK(COP_PREFIX_LDINTERFFTN);
 VMCASE(COP_PREFIX_PACK_VARARGS):
 {
 	/* Pack a set of arguments for a vararg method call */
-#ifdef IL_CONFIG_VARARGS
+#ifdef IL_CONFIG_VARARGS	
+
 	COPY_STATE_TO_THREAD();
 	tempNum = _ILPackVarArgs(thread, stacktop, CVMP_ARG_WORD, CVMP_ARG_WORD2,
 						  CVMP_ARG_WORD2_PTR(ILType *), &tempptr);
 	RESTORE_STATE_FROM_THREAD();
+
 	stacktop -= tempNum;
 	stacktop[0].ptrValue = tempptr;
 	MODIFY_PC_AND_STACK(CVMP_LEN_WORD2_PTR, 1);
@@ -2016,6 +2051,8 @@ VMBREAK(COP_PREFIX_WADDR_NATIVE_N);
 VMCASE(COP_PREFIX_TRACE_IN):
 {
 #if !defined(IL_CONFIG_REDUCE_CODE) && !defined(IL_WITHOUT_TOOLS)
+	BEGIN_NATIVE_CALL();
+
 	ILMutexLock(globalTraceMutex);
 #ifdef INDENT_TRACE
 	int depth = thread->numFrames;
@@ -2032,6 +2069,8 @@ VMCASE(COP_PREFIX_TRACE_IN):
 	putc('\n',stdout);
 	fflush(stdout);
 	ILMutexUnlock(globalTraceMutex);
+
+	END_NATIVE_CALL();
 #endif
 	MODIFY_PC_AND_STACK(CVMP_LEN_WORD, 0);
 }
@@ -2060,7 +2099,11 @@ VMCASE(COP_PREFIX_TRACE_OUT):
 #if !defined(IL_CONFIG_REDUCE_CODE) && !defined(IL_WITHOUT_TOOLS)
 	// NOTE: at some point of time use the Reason parameter to
 	// carry more information about this 
-	ILMethod * methodToReturn = ILExecThreadStackMethod(thread, 1);
+	ILMethod * methodToReturn;
+
+	BEGIN_NATIVE_CALL();	
+
+	methodToReturn = ILExecThreadStackMethod(thread, 1);
 	ILMutexLock(globalTraceMutex);
 #ifdef INDENT_TRACE
 	int depth = thread->numFrames;
@@ -2084,6 +2127,8 @@ VMCASE(COP_PREFIX_TRACE_OUT):
 	putc('\n',stdout);
 	fflush(stdout);
 	ILMutexUnlock(globalTraceMutex);
+
+	END_NATIVE_CALL();
 #endif
 	MODIFY_PC_AND_STACK(CVMP_LEN_WORD, 0);
 }
