@@ -27,17 +27,56 @@
 #include "il_coder.h"
 #include "il_align.h"
 #include "il_gc.h"
+#include "il_utils.h"
 #include "cvm.h"
+#include "interrupt.h"
 
 #ifdef	__cplusplus
 extern	"C" {
 #endif
 
+/* State of an ILExecProcess/AppDomain */
+#define _IL_PROCESS_STATE_UNLOADING (1)
+#define _IL_PROCESS_STATE_UNLOADED  (2)
+
+/* Flags that represents various tasks that need to be performed
+   at safe points */
+#define _IL_MANAGED_SAFEPOINT_THREAD_ABORT		(1)
+#define _IL_MANAGED_SAFEPOINT_THREAD_SUSPEND	(2)
+
+/* IL_SETJMP return value for null pointer interrupts */
+#define _IL_INTERRUPT_NULL_POINTER	(-1)
+#define _IL_INTERRUPT_INT_DIVIDE_BY_ZERO (-2)
+#define _IL_INTERRUPT_INT_OVERFLOW (-3)
+
+/* Determine the interrupts we should catch */
+
+#if defined(IL_INTERRUPT_HAVE_X86_CONTEXT) || !defined(__GNUC__) || defined(IL_NO_ASM)
+
+#if defined(IL_INTERRUPT_SUPPORTS_ILLEGAL_MEMORY_ACCESS)	
+	#define IL_USE_INTERRUPT_BASED_X (1)
+	#define IL_USE_INTERRUPT_BASED_NULL_POINTER_CHECKS (1)
+#endif
+
+#if defined(IL_INTERRUPT_SUPPORTS_INT_DIVIDE_BY_ZERO)	
+	#define IL_USE_INTERRUPT_BASED_X (1)
+	#define IL_USE_INTERRUPT_BASED_INT_DIVIDE_BY_ZERO_CHECKS (1)
+#endif
+
+#if defined(IL_INTERRUPT_SUPPORTS_INT_OVERFLOW)
+	#define IL_USE_INTERRUPT_BASED_X (1)
+	#define IL_USE_INTERRUPT_BASED_INT_OVERFLOW_CHECKS (1)
+#endif
+
+#endif
+
 /*
  * Determine if this compiler supports inline functions.
  */
 #ifdef __GNUC__
 	#define	IL_INLINE	__inline__
+#elif defined(_MSC_VER)
+	#define IL_INLINE __forceinline
 #else
 	#define	IL_INLINE
 #endif
@@ -135,6 +174,9 @@ struct _tagILExecProcess
 	/* The coder in use by this process */
 	ILCoder		   *coder;
 
+	/* State of the process */
+	int				state;
+
 	/* Useful builtin classes */
 	ILClass        *objectClass;
 	ILClass        *stringClass;
@@ -204,7 +246,6 @@ struct _tagILExecProcess
 	ILUInt32			imtBase;
 
 #endif
-
 };
 
 /*
@@ -255,8 +296,10 @@ struct _tagILExecThread
 
 	/* Flag that indicates whether a thread is aborting */
 	volatile int	aborting;
-	/* Flag that indicates whether an abort has been requested */
-	volatile int	abortRequested;
+
+	/* Flag of tasks that need to be performed at safe points in 
+	   managed code */	   
+	volatile int	managedSafePointFlags;
 
 	/* System.Threading.Thread object */
 	ILObject *clrThread;
@@ -295,6 +338,14 @@ struct _tagILExecThread
 
 	/* The frame where the ThreadAbortException was first noticed */
 	ILUInt32		abortHandlerFrame;
+
+#if defined(IL_INTERRUPT_SUPPORTS_ILLEGAL_MEMORY_ACCESS)
+	/* Context for the current interrupt */
+	ILInterruptContext	interruptContext;
+
+	/* Stores state information for interrupt based exception handling */
+	IL_JMP_BUFFER	exceptionJumpBuffer;
+#endif
 };
 
 /*
@@ -428,17 +479,6 @@ int _ILCallMethod(ILExecThread *thread, ILMethod *method,
  */
 int _ILCVMInterpreter(ILExecThread *thread);
 
-/*
- * Inner version of "ILExecThreadCreate".  The caller is
- * responsible for creating the OS-level thread.
- */
-ILExecThread *_ILExecThreadCreate(ILExecProcess *process);
-
-/*
- *	Destroy an engine level thread.
- */
-void _ILExecThreadDestroy(ILExecThread *thread);
-
 /*
  * Lay out a class's fields, virtual methods, and interfaces.
  * Returns zero if there is something wrong with the definition.
@@ -602,6 +642,13 @@ ILCallFrame *_ILGetNextCallFrame(ILExecThread *thread, ILCallFrame *frame);
  */
 ILCallFrame *_ILAllocCallFrame(ILExecThread *thread);
 
+/*
+ * Set a reference type field of a an object instance.
+ */
+int _ILSystemObjectSetField(ILExecThread *thread, ILObject* obj,
+							   const char *fieldName, const char *signature,
+							   ILObject *value);
+
 /*
  * Create a system exception object of a particular class.
  * We do this very carefully, to avoid re-entering the engine.
@@ -699,20 +746,38 @@ ILObject *_ILCustomToObject(ILExecThread *thread, void *ptr,
  */
 ILObject *ILExecThreadGetClrThread(ILExecThread *thread);
 
-/*
- * Associates a support thread with an engine thread and the engine
- * thread with the support thread.
- */
-void _ILThreadExecuteOn(ILThread *thread, ILExecThread *execThread);
+/* Create an ILExecThread */
+ILExecThread *_ILExecThreadCreate(ILExecProcess *process, int ignoreProcessState);
+
+/* Destroy an ILExecThread */
+void _ILExecThreadDestroy(ILExecThread *thread);
+
+typedef struct __tagILThreadExecContext
+{
+	ILExecThread *execThread;
+}
+ILThreadExecContext;
 
 /*
- * Disassociates a support thread with an engine thread and the
- * engine thread with the support thread.
- * The support thread will no longer be able to execute managed
- * code and the engine thread will no longer be able to be used
- * to execute anything until it is reassociated using ILThreadExecuteOn.
+ * Sets and saves the current execution context for the given ILThread.
  */
-void _ILThreadUnexecuteOn(ILThread *thread, ILExecThread *execThread);
+void _ILThreadSetExecContext(ILThread *thread, ILThreadExecContext *context,
+			ILThreadExecContext *saveContext);
+
+/*
+ * Sets and saves the current execution context for the given ILThread.
+ */
+void _ILThreadSaveExecContext(ILThread *thread, ILThreadExecContext *saveContext);
+
+/*
+ * Restores the execution context for the given ILThread.
+ */
+void _ILThreadRestoreExecContext(ILThread *thread, ILThreadExecContext *context);
+
+/*
+ * Clears the execution context for the given ILThread.
+ */
+void _ILThreadClearExecContext(ILThread *thread);
 
 /*
  *	Constructs and returns a new ThreadAbortException for the given thread.
@@ -725,14 +790,26 @@ ILObject *_ILExecThreadNewThreadAbortException(ILExecThread *thread, ILObject *s
 int _ILExecThreadCurrentExceptionThreadIsAbortException(ILExecThread *thread);
 
 /*
- *	Aborts the current thread.
+ * Suspend the given thread.
  */
-void _ILAbortThread(ILExecThread *thread);
+void _ILExecThreadSuspendThread(ILExecThread *thread, ILThread *supportThread);
 
 /*
- *	Handles thread aborts & interruption.
+ * Abort the given thread.
  */
-void _ILHandleWaitResult(ILExecThread *thread, int result);
+void _ILExecThreadAbortThread(ILExecThread *thread, ILThread *supportThread);
+
+/*
+ * Called by the current thread when it was to begin its abort sequence.
+ * Returns 0 if the thread has successfully self aborted.
+ */
+int _ILExecThreadSelfAborting(ILExecThread *thread);
+
+/*
+ * Handles thread aborts & interruption.
+ * Returns the result.
+ */
+int _ILExecThreadHandleWaitResult(ILExecThread *thread, int result);
 
 /*
  * Creates a monitor used by the execution engine.
