@@ -79,7 +79,7 @@ void _IL_GC_SuppressFinalize(ILExecThread *_thread, ILObject *obj)
  */
 void _IL_GC_WaitForPendingFinalizers(ILExecThread *_thread)
 {	
-	ILGCInvokeFinalizers();
+	_ILExecThreadHandleWaitResult(_thread, ILGCInvokeFinalizers(-1));
 }
 
 /*
@@ -172,6 +172,10 @@ ILNativeInt _IL_GCHandle_GCAddrOfPinnedObject(ILExecThread *_thread,
 			else
 			{
 				object = GetObjectFromGcBase(ptr);
+				if(_ILIsSArray((System_Array *)object))
+				{
+					object = ((void *)(((System_Array *)(object)) + 1));
+				}
 			}
 		}
 	}
