@@ -163,7 +163,7 @@ static ILObject *DeserializeObject(ILExecThread *thread,
 					break;
 			case IL_META_SERIALTYPE_STRING:
 					strLen = ILSerializeReaderGetString(reader,&strValue);
-					if(strLen == -1) 
+					if(strLen == -1 || !strValue) 
 					{
 						return NULL;
 					}
@@ -171,7 +171,7 @@ static ILObject *DeserializeObject(ILExecThread *thread,
 									strValue,strLen));	
 			case IL_META_SERIALTYPE_TYPE:
 					strLen = ILSerializeReaderGetString(reader,&strValue);
-					if(strLen == -1) 
+					if(strLen == -1 || !strValue) 
 					{
 						return NULL;
 					}
@@ -233,7 +233,7 @@ static ILObject *DeserializeObject(ILExecThread *thread,
 						{
 							strLen = ILSerializeReaderGetString(reader,
 															&strValue);
-							if(strLen == -1) 
+							if(strLen == -1 || !strValue) 
 							{
 								return NULL;
 							}
@@ -884,8 +884,44 @@ ILObject *_IL_Assembly_GetExecutingAssembly(ILExecThread *thread)
 System_Array *_IL_Assembly_GetExportedTypes(ILExecThread *_thread,
 											ILObject *_this)
 {
-	/* TODO */
-	return 0;
+	/* TODO: Figure out exactly what is an exported type , right now
+	         this method returns all types in the assembly
+	*/
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILImage *image = ((item != 0) ? ILProgramItem_Image(item) : 0);
+	System_Array *array=NULL; 
+	ILObject **buffer=NULL;   
+	ILUInt32 num=0; 
+	ILClass *classInfo=NULL;
+
+	if(item && _ILClrCheckItemAccess(_thread, item)) 
+	{
+		num = ILImageNumTokens (image, IL_META_TOKEN_TYPE_DEF);
+		array = (System_Array *)ILExecThreadNew(_thread, "[oSystem.Object;",
+												"(Ti)V", (ILVaInt)num);
+		if(!array)
+		{
+			return 0;
+		}
+		buffer = (ILObject **)(ArrayToBuffer(array));
+  		while ((classInfo = (ILClass *) ILImageNextToken 
+							(image, IL_META_TOKEN_TYPE_DEF,classInfo)) != 0)
+		{
+			if (classInfo)
+			{
+				*buffer = _ILGetClrType(_thread,classInfo);
+				if(!(*buffer)) //error getting type
+				{
+					return 0;
+				}
+				++buffer;
+			}
+		}
+		return array;
+	}
+	/* Invalid item, or insufficient access: return a zero-element array */
+	return (System_Array *)ILExecThreadNew
+				(_thread, "[oSystem.Object;", "(Ti)V", (ILVaInt)0);
 }
 
 /*
@@ -2748,6 +2784,7 @@ static ILObject *InvokeMethod(ILExecThread *thread, ILMethod *method,
 					return 0;
 				}
 
+				objectType = 0;
 				args[argNum].objValue = 0;
 			}
 
@@ -3213,7 +3250,43 @@ ILObject *_IL_Module_GetType(ILExecThread *_thread, ILObject *_this,
  */
 System_Array *_IL_Module_GetTypes(ILExecThread *_thread, ILObject *_this)
 {
-	/* TODO */
+	ILProgramItem *item = (ILProgramItem *)_ILClrFromObject(_thread, _this);
+	ILModule *module = ((item != 0) ? ILProgramItemToModule(item) : 0);
+	System_Array *array=NULL; 
+	ILObject **buffer=NULL;   
+	ILUInt32 num=0; 
+	ILClass *classInfo=NULL;
+
+	if(module && _ILClrCheckItemAccess(_thread, item)) 
+	{
+		ILImage *image = ILProgramItem_Image(module);
+		num = ILImageNumTokens (image, IL_META_TOKEN_TYPE_DEF);
+		array = (System_Array *)ILExecThreadNew(_thread, "[oSystem.Object;",
+												"(Ti)V", (ILVaInt)num);
+		if(!array)
+		{
+			return 0;
+		}
+		buffer = (ILObject **)(ArrayToBuffer(array));
+  		while ((classInfo = (ILClass *) ILImageNextToken 
+							(image, IL_META_TOKEN_TYPE_DEF,classInfo)) != 0)
+		{
+			if (classInfo)
+			{
+				*buffer = _ILGetClrType(_thread,classInfo);
+				if(!(*buffer)) //error getting type
+				{
+					return 0;
+				}
+				++buffer;
+			}
+		}
+		return array;
+	}
+	else
+	{
+		/* TODO: metadata-based file modules */
+	}
 	return 0;
 }
 
