@@ -48,11 +48,6 @@ extern	"C" {
  */
 #define IL_CLR_WAIT_TIMEOUT (258L)
 
-/*
- * Maximum number of wait handles that can be processed at once.
- */
-#define	IL_MAX_WAIT_HANDLES		64
-
 
 /*************************************/
 /* Interlocked functions for ILInt32 */
@@ -208,7 +203,6 @@ ILObject *_IL_Interlocked_CompareExchange_RObjectObjectObject
 	return (ILObject *)ILInterlockedCompareAndExchangePointers((void **)location1, (void *)value, (void *)comparand);
 }
 
-
 /*
  *  Thread function for ILThread.
  *
@@ -220,7 +214,7 @@ static void __PrivateThreadStart(void *objectArg)
 	ILExecValue result;
 	ILExecThread *thread;
 
-	thread = (ILExecThread *)objectArg;
+	thread = ILExecThreadCurrent();
 
 	/* Get the ThreadStart.Invoke method */
 	method = ILExecThreadLookupMethod(thread, "System.Threading.ThreadStart", "Invoke", "(T)V");
@@ -234,6 +228,8 @@ static void __PrivateThreadStart(void *objectArg)
 	{				
 		ILExecThreadPrintException(thread);
 	}
+
+	ILThreadClearStack(4096);
 }
 
 /*
@@ -306,38 +302,7 @@ void _IL_Thread_FinalizeThread(ILExecThread *thread, ILObject *_this)
  */
 void _IL_Thread_Abort(ILExecThread *thread, ILObject *_this)
 {
-	ILThread *supportThread;
-	ILExecThread *execThread;
-
-	supportThread = ((System_Thread *)_this)->privateData;
-
-	/* Lock the process to prevent the ILExecThread from being destroyed
-	   while we are accessing it (This could happen if Abort is called on a
-	   thread that has finished and is in the process of cleaning up itself) */
-	
-	ILMutexLock(thread->process->lock);
-
-	execThread = ILExecThreadFromThread(supportThread);
-
-	if (execThread == 0)
-	{		
-		ILMutexUnlock(thread->process->lock);
-		return;
-	}
-	else
-	{
-		execThread->abortRequested = 1;
-	}
-
-	ILMutexUnlock(thread->process->lock);
-
-	ILThreadAbort(supportThread);
-
-	/* If the current thread is aborting itself then abort immediately */
-	if (supportThread == thread->supportThread)
-	{
-		_ILAbortThread(thread);		
-	}
+	_ILExecThreadAbortThread(thread, ((System_Thread *)_this)->privateData);
 }
 
 /*
@@ -355,16 +320,8 @@ void _IL_Thread_Interrupt(ILExecThread *thread, ILObject *_this)
  * public void Suspend();
  */
 void _IL_Thread_Suspend(ILExecThread *thread, ILObject *_this)
-{	
-	if (ILThreadSuspend(((System_Thread *)_this)->privateData) == 0)
-	{
-		ILExecThreadThrowSystem
-			(
-				thread,
-				"System.Threading.ThreadStateException",
-				(const char *)0
-			);
-	}
+{
+	_ILExecThreadSuspendThread(thread, ((System_Thread *)_this)->privateData);
 }
 
 /*
@@ -394,7 +351,10 @@ void _IL_Thread_SpinWait(ILExecThread *_thread, ILInt32 iterations)
 		 *
 		 * Also, usleep() isn't thread safe on some platforms.
 		 */
-		ILThreadYield();		
+		while ((iterations--) > 0)
+		{
+			ILThreadYield();
+		}
 	}
 }
 
@@ -439,7 +399,7 @@ ILBool _IL_Thread_InternalJoin(ILExecThread *thread, ILObject *_this,
 
 	case IL_JOIN_ABORTED:
 
-		_ILAbortThread(thread);
+		_ILExecThreadSelfAborting(thread);
 
 		return 0;
 
@@ -481,8 +441,11 @@ void _IL_Thread_ResetAbort(ILExecThread *thread)
 
 	if (ILThreadAbortReset())
 	{
+		ILThreadAtomicStart();
+		thread->managedSafePointFlags &= ~_IL_MANAGED_SAFEPOINT_THREAD_ABORT;
+		ILThreadAtomicEnd();
+
 		thread->aborting = 0;
-		thread->abortRequested = 0;
 		thread->abortHandlerEndPC = 0;
 		thread->abortHandlerFrame = 0;
 		thread->threadAbortException = 0;
@@ -494,17 +457,13 @@ void _IL_Thread_ResetAbort(ILExecThread *thread)
  */
 void _IL_Thread_InternalSleep(ILExecThread *thread, ILInt32 timeout)
 {
-	int result;
-	
 	if (timeout < -1)
 	{
 		ILExecThreadThrowSystem(thread,
 			"System.ArgumentOutOfRangeException", (const char *)0);
 	}
 	
-	result = ILThreadSleep((ILUInt32)timeout);
-	
-	_ILHandleWaitResult(thread, result);
+	_ILExecThreadHandleWaitResult(thread, ILThreadSleep((ILUInt32)timeout));
 }
 
 /*
@@ -512,22 +471,25 @@ void _IL_Thread_InternalSleep(ILExecThread *thread, ILInt32 timeout)
  */
 void _IL_Thread_Start(ILExecThread *thread, ILObject *_this)
 {
+	ILWaitHandle *monitor;
 	ILThread *supportThread;
 	ILExecThread *execThread;
 
-	_IL_Monitor_Enter(thread, _this);
-
 	/* Get the support thread stored inside the first field of the 
 	CLR thread by _IL_Thread_InitializeThread */
 
 	supportThread = ((System_Thread *)_this)->privateData;
+
+	monitor = ILThreadGetMonitor(supportThread);
+
+	ILWaitMonitorEnter(monitor);
 	
 	if (supportThread == 0
 		|| (ILThreadGetState(supportThread) & IL_TS_UNSTARTED) == 0)
 	{
 		/* Thread has already been started or has ended. */
 
-		_IL_Monitor_Exit(thread, _this);
+		ILWaitMonitorLeave(monitor);
 
 		ILExecThreadThrowSystem(thread, "System.Threading.ThreadStateException", 
 			"Exception_ThreadAlreadyStarted");
@@ -539,12 +501,12 @@ void _IL_Thread_Start(ILExecThread *thread, ILObject *_this)
 
 	if ((execThread = ILThreadRegisterForManagedExecution(thread->process, supportThread)) == 0)
 	{
-		/* Runtime isn't fully initialized.  This should never happen */
+		ILWaitMonitorLeave(monitor);
 
-		_IL_Monitor_Exit(thread, _this);
-
-		ILExecThreadThrowSystem(thread, "System.ExecutionEngineException",
-			"Exception_UnexpectedEngineState");
+		if ((thread->process->state & (_IL_PROCESS_STATE_UNLOADED | _IL_PROCESS_STATE_UNLOADING)) == 0)
+		{
+			ILExecThreadThrowOutOfMemory(thread);
+		}
 		
 		return;
 	}
@@ -557,10 +519,11 @@ void _IL_Thread_Start(ILExecThread *thread, ILObject *_this)
 
 	if (ILThreadStart(supportThread) == 0)
 	{
-		/* Start unsuccessul.  Do a GC then try again */
+		/* Start unsuccessul.  Do a GC then try again. */
 
 		ILGCCollect();
-		ILGCInvokeFinalizers();
+		/* Wait a resonable amount of time (1 sec) for finalizers to run */
+		ILGCInvokeFinalizers(1000);
 	
 		if (ILThreadStart(supportThread) == 0)
 		{
@@ -575,7 +538,7 @@ void _IL_Thread_Start(ILExecThread *thread, ILObject *_this)
 		}
 	}
 	
-	_IL_Monitor_Exit(thread, _this);
+	ILWaitMonitorLeave(monitor);
 }
 
 /*
@@ -583,7 +546,7 @@ void _IL_Thread_Start(ILExecThread *thread, ILObject *_this)
  */
 ILObject *_IL_Thread_InternalCurrentThread(ILExecThread *thread)
 {
-	return ILExecThreadGetClrThread(thread);
+	return ILExecThreadCurrentClrThread();
 }
 
 /*
@@ -659,16 +622,24 @@ ILBool _IL_Thread_CanStartThreads(ILExecThread *_thread)
 	return ILHasThreads();
 }
 
+/*
+ * The Thread.Volatile methods guarantee not only ordering but
+ * correct correct flused values on multiprocessor systems which 
+ * is why memory barriers are required on each call.  The C volatile
+ * keyword is far from sufficient and most likely has no effect
+ * but it can't hurt to use them anyway :D.
+ *
+ * -Tum
+ */
+
 /*
  * public static sbyte VolatileRead(ref sbyte address);
  */
 ILInt8 _IL_Thread_VolatileRead_Rb(ILExecThread *thread, ILInt8 *address)
 {
-	ILInt8 value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+	
+	return *((volatile ILInt8 *)address);
 }
 
 /*
@@ -676,11 +647,9 @@ ILInt8 _IL_Thread_VolatileRead_Rb(ILExecThread *thread, ILInt8 *address)
  */
 ILUInt8 _IL_Thread_VolatileRead_RB(ILExecThread *thread, ILUInt8 *address)
 {
-	ILUInt8 value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+	
+	return *((volatile ILUInt8 *)address);
 }
 
 /*
@@ -688,11 +657,9 @@ ILUInt8 _IL_Thread_VolatileRead_RB(ILExecThread *thread, ILUInt8 *address)
  */
 ILInt16 _IL_Thread_VolatileRead_Rs(ILExecThread *thread, ILInt16 *address)
 {
-	ILInt16 value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILInt16 *)address);
 }
 
 /*
@@ -700,11 +667,9 @@ ILInt16 _IL_Thread_VolatileRead_Rs(ILExecThread *thread, ILInt16 *address)
  */
 ILUInt16 _IL_Thread_VolatileRead_RS(ILExecThread *thread, ILUInt16 *address)
 {
-	ILUInt16 value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+	
+	return *((volatile ILUInt16 *)address);
 }
 
 /*
@@ -712,11 +677,9 @@ ILUInt16 _IL_Thread_VolatileRead_RS(ILExecThread *thread, ILUInt16 *address)
  */
 ILInt32 _IL_Thread_VolatileRead_Ri(ILExecThread *thread, ILInt32 *address)
 {
-	ILInt32 value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILInt32 *)address);
 }
 
 /*
@@ -724,11 +687,9 @@ ILInt32 _IL_Thread_VolatileRead_Ri(ILExecThread *thread, ILInt32 *address)
  */
 ILUInt32 _IL_Thread_VolatileRead_RI(ILExecThread *thread, ILUInt32 *address)
 {
-	ILUInt32 value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILUInt32 *)address);
 }
 
 /*
@@ -736,11 +697,9 @@ ILUInt32 _IL_Thread_VolatileRead_RI(ILExecThread *thread, ILUInt32 *address)
  */
 ILInt64 _IL_Thread_VolatileRead_Rl(ILExecThread *thread, ILInt64 *address)
 {
-	ILInt64 value;
-	ILThreadAtomicStart();
-	ILMemCpy(&value, address, sizeof(ILInt64));
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILInt64 *)address);
 }
 
 /*
@@ -748,11 +707,9 @@ ILInt64 _IL_Thread_VolatileRead_Rl(ILExecThread *thread, ILInt64 *address)
  */
 ILUInt64 _IL_Thread_VolatileRead_RL(ILExecThread *thread, ILUInt64 *address)
 {
-	ILUInt64 value;
-	ILThreadAtomicStart();
-	ILMemCpy(&value, address, sizeof(ILUInt64));
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILUInt64 *)address);
 }
 
 /*
@@ -761,11 +718,9 @@ ILUInt64 _IL_Thread_VolatileRead_RL(ILExecThread *thread, ILUInt64 *address)
 ILNativeInt _IL_Thread_VolatileRead_Rj(ILExecThread *thread,
 									   ILNativeInt *address)
 {
-	ILNativeInt value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILNativeInt *)address);
 }
 
 /*
@@ -774,11 +729,9 @@ ILNativeInt _IL_Thread_VolatileRead_Rj(ILExecThread *thread,
 ILNativeUInt _IL_Thread_VolatileRead_RJ(ILExecThread *thread,
 										ILNativeUInt *address)
 {
-	ILNativeUInt value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILNativeUInt *)address);
 }
 
 /*
@@ -786,11 +739,9 @@ ILNativeUInt _IL_Thread_VolatileRead_RJ(ILExecThread *thread,
  */
 ILFloat _IL_Thread_VolatileRead_Rf(ILExecThread *thread, ILFloat *address)
 {
-	ILFloat value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILFloat *)address);
 }
 
 /*
@@ -798,11 +749,9 @@ ILFloat _IL_Thread_VolatileRead_Rf(ILExecThread *thread, ILFloat *address)
  */
 ILDouble _IL_Thread_VolatileRead_Rd(ILExecThread *thread, ILDouble *address)
 {
-	ILDouble value;
-	ILThreadAtomicStart();
-	ILMemCpy(&value, address, sizeof(ILDouble));
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return *((volatile ILDouble *)address);
 }
 
 /*
@@ -811,11 +760,9 @@ ILDouble _IL_Thread_VolatileRead_Rd(ILExecThread *thread, ILDouble *address)
 ILObject *_IL_Thread_VolatileRead_RObject(ILExecThread *thread,
 										  ILObject **address)
 {
-	ILObject *value;
-	ILThreadAtomicStart();
-	value = *address;
-	ILThreadAtomicEnd();
-	return value;
+	ILInterlockedMemoryBarrier();
+
+	return (ILObject *)(*((ILObject volatile * volatile *)address));
 }
 
 /*
@@ -824,9 +771,9 @@ ILObject *_IL_Thread_VolatileRead_RObject(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rbb(ILExecThread *thread,
 								  ILInt8 *address, ILInt8 value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILInt8 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -835,9 +782,9 @@ void _IL_Thread_VolatileWrite_Rbb(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_RBB(ILExecThread *thread,
 								  ILUInt8 *address, ILUInt8 value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILUInt8 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -846,9 +793,9 @@ void _IL_Thread_VolatileWrite_RBB(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rss(ILExecThread *thread,
 								  ILInt16 *address, ILInt16 value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILInt16 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -857,9 +804,9 @@ void _IL_Thread_VolatileWrite_Rss(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_RSS(ILExecThread *thread,
 								  ILUInt16 *address, ILUInt16 value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILUInt16 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -868,9 +815,9 @@ void _IL_Thread_VolatileWrite_RSS(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rii(ILExecThread *thread,
 								  ILInt32 *address, ILInt32 value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILInt32 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -879,9 +826,9 @@ void _IL_Thread_VolatileWrite_Rii(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_RII(ILExecThread *thread,
 								  ILUInt32 *address, ILUInt32 value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILUInt32 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -890,9 +837,9 @@ void _IL_Thread_VolatileWrite_RII(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rll(ILExecThread *thread,
 								  ILInt64 *address, ILInt64 value)
 {
-	ILThreadAtomicStart();
-	ILMemCpy(address, &value, sizeof(ILInt64));
-	ILThreadAtomicEnd();
+	*((volatile ILInt64 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -901,9 +848,9 @@ void _IL_Thread_VolatileWrite_Rll(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_RLL(ILExecThread *thread,
 								  ILUInt64 *address, ILUInt64 value)
 {
-	ILThreadAtomicStart();
-	ILMemCpy(address, &value, sizeof(ILUInt64));
-	ILThreadAtomicEnd();
+	*((volatile ILUInt64 *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -912,9 +859,9 @@ void _IL_Thread_VolatileWrite_RLL(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rjj(ILExecThread *thread,
 								  ILNativeInt *address, ILNativeInt value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILNativeInt *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -923,9 +870,9 @@ void _IL_Thread_VolatileWrite_Rjj(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_RJJ(ILExecThread *thread,
 								  ILNativeUInt *address, ILNativeUInt value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILNativeUInt *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -934,9 +881,9 @@ void _IL_Thread_VolatileWrite_RJJ(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rff(ILExecThread *thread,
 								  ILFloat *address, ILFloat value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILFloat *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -945,9 +892,9 @@ void _IL_Thread_VolatileWrite_Rff(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_Rdd(ILExecThread *thread,
 								  ILDouble *address, ILDouble value)
 {
-	ILThreadAtomicStart();
-	ILMemCpy(address, &value, sizeof(ILDouble));
-	ILThreadAtomicEnd();
+	*((volatile ILDouble *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -956,9 +903,9 @@ void _IL_Thread_VolatileWrite_Rdd(ILExecThread *thread,
 void _IL_Thread_VolatileWrite_RObjectObject(ILExecThread *thread,
 											ILObject **address, ILObject *value)
 {
-	ILThreadAtomicStart();
-	*address = value;
-	ILThreadAtomicEnd();
+	*((volatile ILObject * volatile *)address) = value;
+
+	ILInterlockedMemoryBarrier();
 }
 
 /*
@@ -981,64 +928,6 @@ void _IL_WaitHandle_InternalClose(ILExecThread *_thread,
 	}
 }
 
-void _ILAbortThread(ILExecThread *thread)
-{
-	/* Determine if we currently have an abort in progress,
-	or if we need to throw a new abort exception */
-	if(!ILThreadIsAborting())
-	{
-		if(ILThreadSelfAborting())
-		{
-			/* Allocate an instance of "ThreadAbortException" and throw */
-
-			ILObject *exception;
-
-			exception = _ILExecThreadNewThreadAbortException(thread, 0);
-			thread->aborting = 1;
-			thread->abortRequested = 0;
-			thread->abortHandlerEndPC = 0;
-			thread->threadAbortException = 0;
-			thread->abortHandlerFrame = 0;
-
-			ILExecThreadThrow(thread, exception);
-		}
-	}
-}
-
-/*
- * Handle the result from an "ILWait*" function call.
- */
-void _ILHandleWaitResult(ILExecThread *thread, int result)
-{
-	switch (result)
-	{
-		case IL_WAIT_INTERRUPTED:
-		{
-			ILExecThreadThrowSystem
-				(
-					thread,
-					"System.Threading.ThreadInterruptedException",
-					(const char *)0
-				);
-		}
-		break;
-		case IL_WAIT_ABORTED:
-		{
-			_ILAbortThread(thread);			
-		}
-		break;
-		case IL_WAIT_FAILED:
-		{
-			ILExecThreadThrowSystem
-				(
-					thread,
-					"System.Threading.SystemException",
-					(const char *)0
-				);
-		}
-	}
-}
-
 /*
  * private static bool InternalWaitAll(WaitHandle[] waitHandles,
  *									   int timeout, bool exitContext);
@@ -1069,7 +958,7 @@ ILBool _IL_WaitHandle_InternalWaitAll(ILExecThread *_thread,
 	}
 	else if (result < 0)
 	{
-		_ILHandleWaitResult(_thread, result);
+		_ILExecThreadHandleWaitResult(_thread, result);
 
 		return 0;
 	}
@@ -1108,7 +997,7 @@ ILInt32 _IL_WaitHandle_InternalWaitAny(ILExecThread *_thread,
 	}
 	else if (result < 0)
 	{
-		_ILHandleWaitResult(_thread, result);
+		_ILExecThreadHandleWaitResult(_thread, result);
 
 		result = 0;
 	}
@@ -1126,7 +1015,7 @@ ILBool _IL_WaitHandle_InternalWaitOne(ILExecThread *_thread,
 	if(privateData)
 	{
 		int result = ILWaitOne((ILWaitHandle *)privateData, timeout);
-		_ILHandleWaitResult(_thread, result);
+		_ILExecThreadHandleWaitResult(_thread, result);
 		return (result == 0);
 	}
 	return 0;
