@@ -356,14 +356,9 @@ int ILExecProcessAddInternalCallTable(ILExecProcess* process,
 ILExecThread *ILExecThreadCurrent(void);
 
 /*
- * Get an ILExecThread from the given thread.
+ *	Gets the current managed thread object.
  */
-ILExecThread *ILExecThreadFromThread(ILThread *thread);
-
-/*
- *	Gets the managed thread object from an engine thread.
- */
-ILObject *ILExecThreadGetClrThread(ILExecThread *thread);
+ILObject *ILExecThreadCurrentClrThread();
 
 /*
  * Get the process that corresponds to a thread.
@@ -425,6 +420,15 @@ ILField *ILExecThreadLookupField(ILExecThread *thread,
 								 const char *fieldName,
 								 const char *signature);
 
+/*
+ * Look up a field in a  particular class.  Returns NULL if
+ * the field could not be found.
+ */
+ILField *ILExecThreadLookupFieldInClass(ILExecThread *thread,
+								 ILClass *classInfo,
+								 const char *fieldName,
+								 const char *signature);
+
 /*
  * Call a particular method within a thread's context.
  * Returns non-zero if an exception was thrown during
