@@ -62,6 +62,19 @@ int ILLinkerCreateModuleAndAssembly(ILLinker *linker,
 									const char *assemblyKey,
 									int hashAlgorithm);
 
+/*
+ * Set the metadata version string in the assembly's header.
+ * If the parameter is set to NULL, then infer the metadata
+ * version from the "mscorlib" library.
+ */
+void ILLinkerSetMetadataVersion(ILLinker *linker, const char *version,
+								const char *stdLibrary);
+
+/*
+ * Set the culture name on the assembly.
+ */
+void ILLinkerSetCulture(ILLinker *linker, const char *culture);
+
 /*
  * Add a directory to a linker context to search for libraries.
  * Returns zero on error.
@@ -178,6 +191,11 @@ int ILLinkerPerformLink(ILLinker *linker);
  */
 int ILLinkerSetFlags(ILLinker *linker, int flags);
 
+/*
+ * Parse an assembly version string.
+ */
+int ILLinkerParseVersion(ILUInt16 *version, const char *str);
+
 /*
  * Call the linker as if it were an executable with command-line
  * options.  It is really linked into the calling program.
