@@ -537,6 +537,11 @@ int ILTypeIsStringClass(ILType *type);
  */
 int ILTypeIsObjectClass(ILType *type);
 
+/*
+ * Determine if a type is a "system/corlib" class.
+ */
+int ILTypeIsSystemClass(ILType *type, const char *namespace, const char *name);
+
 /*
  * Determine if two types are assignment-compatible within
  * the context of a particular image when assigning a value
@@ -582,6 +587,18 @@ int ILTypeIsDelegate(ILType *type);
  */
 void *ILTypeGetDelegateMethod(ILType *type);
 
+/*
+ * Get the "BeginInvoke" method associated with a delegate type.
+ * Returns NULL if not a delegate type.
+ */
+void *ILTypeGetDelegateBeginInvokeMethod(ILType *type);
+
+/*
+ * Get the "EndInvoke" method associated with a delegate type.
+ * Returns NULL if not a delegate type.
+ */
+void *ILTypeGetDelegateEndInvokeMethod(ILType *type);
+
 /*
  * Determine if we have a signature match between a delegate
  * type and a particular method.
