@@ -599,14 +599,14 @@ static void MulByWord(ILUInt32 *result, int base,
 	--base;
 
 	/* valuea->middle * valueb */
-	product = ((ILUInt64)(valuea->middle)) * ((ILUInt64)valueb);
+	product += ((ILUInt64)(valuea->middle)) * ((ILUInt64)valueb);
 	product += (ILUInt64)(result[base]);
 	result[base] = (ILUInt32)product;
 	product >>= 32;
 	--base;
 
 	/* valuea->high * valueb */
-	product = ((ILUInt64)(valuea->high)) * ((ILUInt64)valueb);
+	product += ((ILUInt64)(valuea->high)) * ((ILUInt64)valueb);
 	product += (ILUInt64)(result[base]);
 	result[base] = (ILUInt32)product;
 	product >>= 32;
@@ -757,7 +757,7 @@ static int Divide(ILUInt32 *quotient, const ILDecimal *valuea,
 	ILUInt32 testquot;
 
 	/* Expand valuea to a 228-bit value, shifted up by 29 places */
-	tempa[0] = tempa[1] = tempa[2] = tempb[3] = 0;
+	tempa[0] = tempa[1] = tempa[2] = tempa[3] = 0;
 	tempa[4] = valuea->high;
 	tempa[5] = valuea->middle;
 	tempa[6] = valuea->low;
