@@ -29,7 +29,15 @@ You normally won't need to modify or replace this file when porting.
 */
 
 #include "thr_defs.h"
-#include "signal.h"
+#include "interlocked.h"
+#include "interrupt.h"
+
+#ifdef HAVE_ALLOCA_H
+#include <alloca.h>
+#endif
+#ifdef HAVE_MALLOC_H
+#include <malloc.h>
+#endif
 
 #ifdef	__cplusplus
 extern	"C" {
@@ -83,7 +91,14 @@ static void _ILThreadInit(void)
 	_ILSemaphoreCreate(&(mainThread.resumeAck));
 	_ILSemaphoreCreate(&(mainThread.suspendAck));
 	mainThread.startFunc        = 0;
-	mainThread.objectArg        = 0;
+	mainThread.userObject       = 0;
+	mainThread.startArg			= 0;
+	mainThread.destroyOnExit	= 0;
+	mainThread.monitor = ILWaitMonitorCreate();
+	#ifdef IL_INTERRUPT_SUPPORTS
+		mainThread.interruptHandler = 0;
+	#endif
+
 	_ILWakeupCreate(&(mainThread.wakeup));
 	_ILWakeupQueueCreate(&(mainThread.joinQueue));
 
@@ -95,14 +110,18 @@ static void _ILThreadInit(void)
 	/* We have 1 foreground thread in the system at present */
 	numThreads = 1;
 	numBackgroundThreads = 0;
+
+	_ILInterruptInit();
 }
 
 static void _ILThreadDeinit(void)
 {
+	_ILInterruptDeinit();
+
 	if(noFgThreadsEvent != 0)
 	{
 		ILWaitHandleClose(noFgThreadsEvent);
-	}
+	}	
 }
 
 /*
@@ -139,6 +158,11 @@ void ILThreadDeinit(void)
 	_ILCallOnce(_ILThreadDeinit);
 }
 
+ILWaitHandle *ILThreadGetMonitor(ILThread *thread)
+{
+	return thread->monitor;
+}
+
 static void _ILThreadRunAndFreeCleanups(ILThread *thread)
 {
 	ILThreadCleanupEntry *entry, *next;
@@ -156,6 +180,70 @@ static void _ILThreadRunAndFreeCleanups(ILThread *thread)
 
 		entry = next;
 	}
+
+	thread->firstCleanupEntry = 0;
+}
+
+static void _ILPrivateThreadDestroy(ILThread *thread, int allowSelf)
+{
+	/* Bail out if this is the current thread or main thread */
+	if((thread == _ILThreadGetSelf() && !allowSelf) || thread == &mainThread)
+	{
+		return;
+	}
+
+	/* Lock down the thread object */
+	_ILMutexLock(&(thread->lock));
+
+	/* Don't terminate the thread or adjust counts if it has already been stopped */
+	if((thread->state & IL_TS_STOPPED) == 0)
+	{
+		thread->state |= IL_TS_STOPPED;
+
+		/* Only terminate the system thread if one was created */
+		if((thread->state & IL_TS_UNSTARTED) == 0)
+		{
+			/* Terminating the thread is unsafe so just don't
+			   destroy now and tell the thread to destroy itself
+			   on exit */
+
+			thread->destroyOnExit = 1;
+
+			_ILMutexUnlock(&thread->lock);
+			
+			ILThreadAbort(thread);
+
+			return;
+		}
+
+		_ILMutexUnlock(&thread->lock);
+
+		/* Run and free the cleanup handlers */
+		_ILThreadRunAndFreeCleanups(thread);
+	}
+	else
+	{
+		/* Unlock the thread object and free it */
+		_ILMutexUnlock(&(thread->lock));
+	}
+
+	/* Only destroy the system thread if one was created */
+	if((thread->state & IL_TS_UNSTARTED) == 0)
+	{
+		_ILThreadDestroy(thread);
+	}
+
+	ILWaitHandleClose(thread->monitor);
+	_ILMutexDestroy(&(thread->lock));
+	_ILSemaphoreDestroy(&(thread->suspendAck));
+	_ILSemaphoreDestroy(&(thread->resumeAck));	
+	_ILWakeupQueueDestroy(&(thread->joinQueue));
+	ILFree(thread);
+}
+
+void ILThreadDestroy(ILThread *thread)
+{
+	_ILPrivateThreadDestroy(thread, 0);
 }
 
 void _ILThreadRun(ILThread *thread)
@@ -169,14 +257,16 @@ void _ILThreadRun(ILThread *thread)
 	   was aborted before it even got going */
 	if(thread->startFunc)
 	{
-		(*(thread->startFunc))(thread->objectArg);
+		(*(thread->startFunc))(thread->startArg);
 	}
 
+	thread->startArg = 0;
+	
 	_ILMutexLock(&(thread->lock));
 	{
 		/* Mark the thread as stopped */
 		thread->state |= IL_TS_STOPPED;	
-		/* Change the thread count */
+		/* Change the thread count */		
 		_ILThreadAdjustCount(-1, ((thread->state & IL_TS_BACKGROUND) != 0) ? -1 : 0);
 	}
 	_ILMutexUnlock(&(thread->lock));
@@ -190,9 +280,18 @@ void _ILThreadRun(ILThread *thread)
 		_ILWakeupQueueWakeAll(&(thread->joinQueue));
 	}
 	_ILMutexUnlock(&(thread->lock));
+
+	/* The wakeup isn't needed anymore so destroy it now to allow
+	   held mutexes to be released */
+	_ILWakeupDestroy(&(thread->wakeup));
+
+	if (thread->destroyOnExit)
+	{
+		_ILPrivateThreadDestroy(thread, 1);
+	}
 }
 
-ILThread *ILThreadCreate(ILThreadStartFunc startFunc, void *objectArg)
+ILThread *ILThreadCreate(ILThreadStartFunc startFunc, void *startArg)
 {
 	ILThread *thread;
 
@@ -208,20 +307,26 @@ ILThread *ILThreadCreate(ILThreadStartFunc startFunc, void *objectArg)
 		return 0;
 	}
 
-	_ILMutexCreate(&(thread->lock));
+	_ILMutexCreate(&(thread->lock));	
 	thread->state = IL_TS_UNSTARTED;
 	thread->resumeRequested = 0;
 	thread->suspendRequested = 0;
 	thread->numLocksHeld = 0;
 	thread->firstCleanupEntry = 0;
 	thread->lastCleanupEntry = 0;
+	thread->monitor = ILWaitMonitorCreate();
 	_ILSemaphoreCreate(&(thread->resumeAck));
 	_ILSemaphoreCreate(&(thread->suspendAck));
 	thread->startFunc = startFunc;
-	thread->objectArg = objectArg;
+	thread->userObject = 0;
+	thread->startArg = startArg;
 	_ILWakeupCreate(&(thread->wakeup));
 	_ILWakeupQueueCreate(&(thread->joinQueue));
 	thread->handle = 0;
+	thread->destroyOnExit = 0;
+	#ifdef IL_INTERRUPT_SUPPORTS
+		thread->interruptHandler = 0;
+	#endif
 
 	/* Lock out the thread system */
 	_ILMutexLock(&threadLockAll);
@@ -253,7 +358,11 @@ int ILThreadStart(ILThread *thread)
 			thread->state |= IL_TS_RUNNING;
 
 			_ILThreadAdjustCount(1, (thread->state & IL_TS_BACKGROUND) ? 1 : 0);
-			
+
+		#if defined(GC_OPENBSD_THREADS)
+			pthread_yield();
+		#endif
+
 			/* Let the thread start running */
 			_ILThreadResumeSelf(thread);
 
@@ -270,56 +379,6 @@ int ILThreadStart(ILThread *thread)
 	return result;
 }
 
-void ILThreadDestroy(ILThread *thread)
-{
-	/* Bail out if this is the current thread or main thread */
-	if(thread == _ILThreadGetSelf() || thread == &mainThread)
-	{
-		return;
-	}
-
-	/* Lock down the thread object */
-	_ILMutexLock(&(thread->lock));
-	
-	/* Don't terminate the thread or adjust counts if it has already been stopped */
-	if((thread->state & IL_TS_STOPPED) == 0)
-	{
-		thread->state |= IL_TS_STOPPED;
-
-		/* Only terminate the system thread if one was created */
-		if((thread->state & IL_TS_UNSTARTED) == 0)
-		{
-			_ILThreadTerminate(thread);
-
-			/* Adjust the thread count */
-			_ILThreadAdjustCount(-1, (thread->state & IL_TS_BACKGROUND) ? -1 : 0);
-		}
-
-		_ILMutexUnlock(&thread->lock);
-		
-		/* Run and free the cleanup handlers */
-		_ILThreadRunAndFreeCleanups(thread);
-	}
-	else
-	{
-		/* Unlock the thread object and free it */
-		_ILMutexUnlock(&(thread->lock));
-	}
-
-	/* Only destroy the system thread if one was created */
-	if((thread->state & IL_TS_UNSTARTED) == 0)
-	{
-		_ILThreadDestroy(thread);
-	}
-
-	_ILMutexDestroy(&(thread->lock));
-	_ILSemaphoreDestroy(&(thread->suspendAck));
-	_ILSemaphoreDestroy(&(thread->resumeAck));
-	_ILWakeupDestroy(&(thread->wakeup));
-	_ILWakeupQueueDestroy(&(thread->joinQueue));
-	ILFree(thread);
-}
-
 ILThread *ILThreadSelf(void)
 {
 #ifdef IL_NO_THREADS
@@ -331,12 +390,12 @@ ILThread *ILThreadSelf(void)
 
 void *ILThreadGetObject(ILThread *thread)
 {
-	return thread->objectArg;
+	return thread->userObject;
 }
 
-void ILThreadSetObject(ILThread *thread, void *objectArg)
+void ILThreadSetObject(ILThread *thread, void *userObject)
 {
-	thread->objectArg = objectArg;
+	thread->userObject = userObject;
 }
 
 int ILThreadSuspend(ILThread *thread)
@@ -349,7 +408,11 @@ int ILThreadSuspend(ILThread *thread)
 
 	/* Determine what to do based on the thread's state */
 	state = thread->state;
-	if((state & IL_TS_SUSPENDED) != 0)
+	if ((state & (IL_TS_ABORT_REQUESTED)) != 0)
+	{
+		return IL_WAIT_ABORTED;
+	}
+	else if((state & IL_TS_SUSPENDED) != 0)
 	{
 		/* Nothing to do here - it is already suspended */
 	}
@@ -439,7 +502,7 @@ void ILThreadResume(ILThread *thread)
 		{
 			/* Someone else suspended the thread */
 			thread->state &= ~IL_TS_SUSPENDED;
-			thread->state |= IL_TS_RUNNING;
+			thread->state |= IL_TS_RUNNING;			
 			_ILThreadResumeOther(thread);
 		}
 	}
@@ -536,6 +599,14 @@ int ILThreadAbort(ILThread *thread)
 						
 			return 0;
 		}
+		else if (((thread->state & (IL_TS_SUSPENDED_SELF | IL_TS_SUSPENDED))) != 0)
+		{
+			_ILMutexUnlock(&(thread->lock));
+
+			ILThreadResume(thread);
+
+			return 0;
+		}
 
 		/* No need to abort the current thread */
 		result = 0;
@@ -795,11 +866,7 @@ void ILThreadAtomicEnd(void)
 
 void ILThreadMemoryBarrier(void)
 {
-	/* For now, we just acquire and release a mutex, under the
-	   assumption that the underlying thread package will do
-	   data synchronisation for us as part of the mutex code */
-	_ILMutexLock(&atomicLock);
-	_ILMutexUnlock(&atomicLock);
+	ILInterlockedMemoryBarrier();
 }
 
 void ILThreadGetCounts(unsigned long *numForeground,
@@ -886,7 +953,7 @@ int ILThreadSleep(ILUInt32 ms)
 		thread->state &= ~IL_TS_SUSPEND_REQUESTED;
 		thread->state |= IL_TS_SUSPENDED | IL_TS_SUSPENDED_SELF;
 		thread->resumeRequested = 0;
-		
+
 		/* Unlock the thread object prior to suspending */
 		_ILMutexUnlock(&(thread->lock));
 
@@ -1047,17 +1114,47 @@ int ILThreadGetPriority(ILThread *thread)
 
 void _ILThreadSuspendRequest(ILThread *thread)
 {
+	_ILMutexLock(&(thread->lock));
+
 	/* Clear the "suspendRequested" and "resumeRequested" flags */
+	thread->state |= IL_TS_SUSPENDED | IL_TS_SUSPENDED_SELF;
 	thread->suspendRequested = 0;
 	thread->resumeRequested = 0;
 
 	/* Signal the thread that wanted to make us suspend that we are */
 	_ILSemaphorePost(&(thread->suspendAck));
 
+	_ILMutexUnlock(&(thread->lock));
+
 	/* Suspend the current thread until we receive a resume signal */
 	_ILThreadSuspendSelf(thread);
 }
 
+/*
+ * Clear stack space.
+ */
+void ILThreadClearStack(int length)
+{
+	char *ptr;
+
+#ifdef HAVE_ALLOCA
+#ifdef _MSC_VER
+	ptr = _alloca(length);
+#else
+	ptr = alloca(length);
+#endif
+#else
+	char stackBuffer[4096];
+	length = 4096;
+	ptr = &stackBuffer[0];
+#endif
+
+	if (ptr)
+	{
+		ILMemZero(ptr, length);
+	}
+}
+
 #ifdef	__cplusplus
 };
 #endif
