@@ -2,6 +2,9 @@
  * time.c - Get the current system time.
  *
  * Copyright (C) 2001  Southern Storm Software, Pty Ltd.
+ * Copyright (C) 2004  Free Software Foundation
+ *
+ * Contributions from Thong Nguyen (tum@veridicus.com)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +22,7 @@
  */
 
 #include "il_system.h"
+#include "il_thread.h"
 #if defined(__palmos__)
 	#include <PalmTypes.h>
 	#include <TimeMgr.h>
@@ -40,6 +44,21 @@
 #define timezone _timezone
 #endif
 
+#if defined(HAVE_SYS_SYSINFO_H) && defined(HAVE_SYSINFO) \
+	&& (defined(linux) \
+	|| defined(__linux) || defined(__linux__))
+	
+	#include <sys/sysinfo.h>
+	#define USE_BOOTTIME 1
+#endif
+
+#if defined(HAVE_SYS_SYSCTL_H) && defined(HAVE_SYSCTL) \
+	&& defined(__FreeBSD__)
+	
+	#include <sys/sysctl.h>
+	#define USE_BOOTTIME 1
+#endif
+
 #ifdef	__cplusplus
 extern	"C" {
 #endif
@@ -92,6 +111,10 @@ void ILGetCurrTime(ILCurrTime *timeValue)
 #endif
 }
 
+#if defined(USE_BOOTTIME)
+static ILCurrTime bootTime;
+#endif
+
 int ILGetSinceRebootTime(ILCurrTime *timeValue)
 {
 #ifdef IL_WIN32_PLATFORM
@@ -103,6 +126,80 @@ int ILGetSinceRebootTime(ILCurrTime *timeValue)
 	timeValue->nsecs = (tick % 1000) * 1000000;
 
 	return 1;
+#elif defined(USE_BOOTTIME) && defined(__FreeBSD__) 
+	int len, mib[2];	
+	struct timeval tv;
+
+	ILGetCurrTime(timeValue);
+
+	if (bootTime.secs == 0 && bootTime.nsecs == 0)
+	{
+		mib[0] = CTL_KERN;
+		mib[1] = KERN_BOOTTIME;
+
+		len = sizeof(struct timeval);
+
+		if (sysctl(mib, 2, &tv, &len, 0, 0) != 0)
+		{
+			return 0;
+		}
+
+		ILThreadAtomicStart();
+
+		bootTime.secs = ((ILInt64)(tv.tv_sec)) + EPOCH_ADJUST;
+		bootTime.nsecs = (ILUInt32)(tv.tv_usec * 1000);
+
+		ILThreadAtomicEnd();
+	}
+	
+#elif defined(USE_BOOTTIME) && (defined(linux) \
+	|| defined(__linux) || defined(__linux__))
+
+	struct sysinfo si;
+	
+	ILGetCurrTime(timeValue);
+
+	if (bootTime.secs == 0 && bootTime.nsecs == 0)
+	{
+		if (sysinfo(&si) != 0)
+		{
+			return 0;
+		}
+		
+		ILThreadAtomicStart();		
+		bootTime.secs = timeValue->secs - si.uptime;
+
+		/* sysinfo() is only accurate to the second so
+		   use the nsec value from the curren time.
+		   This allows subsequent calls to this function
+		   to get nsec time-differential precision  */
+
+		bootTime.nsecs = timeValue->nsecs;
+		ILThreadAtomicEnd();
+	}
+#endif
+
+#if defined(USE_BOOTTIME)
+	/* Subtract the current time from the time since the system
+	   was started */
+
+	if(timeValue->nsecs < bootTime.nsecs)
+	{
+		timeValue->nsecs = 
+			timeValue->nsecs - bootTime.nsecs + 1000000000;
+		timeValue->secs =
+			timeValue->secs - bootTime.secs;
+	}
+	else
+	{
+		timeValue->nsecs =
+			timeValue->nsecs - bootTime.nsecs;			
+		timeValue->secs =
+			timeValue->secs - bootTime.secs;
+	}	
+
+	return 1;
+	
 #else
 	return 0;
 #endif
@@ -127,7 +224,7 @@ ILInt32 ILGetTimeZoneAdjust(void)
 #endif
 		initialized = 1;
 	}
-	return (ILInt32)(timezone - (isdst ? 3600 : 0));
+	return (ILInt32)timezone;
 #else
 	/* TODO */
 	return 0;
